import json

id2handlerInfo = {}


class JSONSerializable:
    def __str__(self):
        return jsonify(self)

    def __repr__(self):
        return jsonify(self, indent=4)


class EventHandlerInfo(JSONSerializable):
    def __init__(self):
        self.selector = None
        self.event = None
        self.uiData = None
        self.handler = None
        self._handler = None
        self.filter_selector = None
        self.stop_propagation = False
        self.throttle = False


def get_handler_info(func) -> EventHandlerInfo:
    function_id = id(func)
    handler_info = id2handlerInfo.get(function_id)
    if handler_info is None:
        handler_info = EventHandlerInfo()
        id2handlerInfo[function_id] = handler_info
    return handler_info


def jsonify(obj, **kwargs) -> str:
    return json.dumps(obj,
                      default=lambda o: {k: v for k, v in o.__dict__.items() if not k.startswith('_')},
                      separators=(',', ':'), **kwargs)


class UIData:
    def __init__(self, data):
        self._data = data

    def __getattr__(self, item):
        return self._data[item]

    def __repr__(self):
        return repr(self._data)



from .decorators import *
import json
import inspect
import websockets

try:
    # noinspection PyPackageRequirements
    from dominate.tags import html_tag
except ImportError:
    # noinspection PyPep8Naming
    class html_tag:
        pass


def preprocess_jquery_arguments(x):
    if isinstance(x, html_tag):
        return str(x)
    if inspect.iscoroutinefunction(x):
        on(None, None)(x)
        return id2handlerInfo[id(x)]
    return x


class Selector(JSONSerializable):
    def __init__(self, selector: str, websocket: websockets.WebSocketServerProtocol=None):
        self.selector = selector
        self.sub_actions = []
        self.type = 'selector'
        self._websocket = websocket
        if websocket is not None:
            self.command = 'immediate'

    def __getattr__(self, action):
        if self._websocket is None:
            def func(*args):
                args = tuple(preprocess_jquery_arguments(a) for a in args)
                self.sub_actions.append((action,) + args)
                return self
            return func

        async def func(*args):
            self.sub_actions.append((action,) + args)
            await self._websocket.send(str(self))
            return json.loads(await self._websocket.recv())
        return func


class Bundle(JSONSerializable):
    def __init__(self, websocket: websockets.WebSocketServerProtocol, command='batch'):
        self.command = command
        self.actions = []
        self._websocket = websocket
        if command == 'batch':
            self.broadcast = Bundle(websocket, command='broadcast')

    def __call__(self, selector: str):
        s = Selector(selector)
        self.actions.append(s)
        return s

    def eval(self, code: str, **data):
        """Run arbitrary code in the browser after the handler returns. 
        Meant to be used for interacting with 3rd-party Javascript libraries or things not provided by jQuery API.

        Args:
            code: The Javascript code to be evaluated in the browser
            **data: The data supplied to be used by the Javascript code, must be JSON serializable
                All data will be attached to the `window` object.
        """
        self.actions.append({'type': 'eval', 'code': code, 'data': data})

    async def eval_immediately(self, code: str, **data):
        """Run arbitrary code in the browser and get the evaluated result back immediately.
        Must be awaited.

        Args:
            code: The Javascript code to be evaluated in the browser
            **data: The data supplied to be used by the Javascript code, must be JSON serializable
                All data will be attached to the `window` object.

        Returns: The evaluated result from browser

        """
        await self._websocket.send(json.dumps({
            'command': 'immediate',
            'type': 'eval',
            'code': code,
            'data': data
        }))
        return json.loads(await self._websocket.recv())

    def immediate(self, selector: str) -> Selector:
        """Execute jQuery function and get the result back immediately.
        Must be awaited.
        Meant to be used for dynamically getting attributes from UI.

        Args:
            selector: A jQuery selector expression

        Returns: The resulting selector object

        """
        return Selector(selector, websocket=self._websocket)



import asyncio
import functools
import logging
import sys
import types

import websockets
# noinspection PyProtectedMember
from werkzeug._reloader import run_with_reloader  # Sorry, I didn't find any easier way to do this.

from .selector import Bundle
from .utilities import *

clients = set()


def _start_app(log_level, **ws_options):
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:  # when auto-reloading, it cannot get event loop in the child thread
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    logger = logging.getLogger('Portkey')
    logger.setLevel(log_level)
    ch = logging.StreamHandler()
    ch.setLevel(log_level)
    ch.setFormatter(logging.Formatter('[%(asctime)s %(levelname)s] %(message)s'))
    logger.addHandler(ch)

    async def socket_handler(ws: websockets.WebSocketServerProtocol, path: str):
        try:
            context = types.SimpleNamespace()
            clients.add(ws)
            logger.info('Connection established on %s.', path)
            client_register_info = jsonify({
                'command': 'register',
                'data': list(id2handlerInfo.values())
            })
            # print('handlerInfo', client_register_info)
            await ws.send(client_register_info)
            logger.info('Client events registered.')
            while 1:
                msg = await ws.recv()
                # print('received', msg)
                msg = json.loads(msg)
                handler_id = msg['handler']
                ui_data = UIData(msg['uiData'])
                selector = Bundle(ws)
                # noinspection PyBroadException
                try:
                    # noinspection PyProtectedMember
                    await id2handlerInfo[handler_id]._handler(selector, ui_data, context)
                except:
                    sys.excepthook(*sys.exc_info())
                broadcast = selector.broadcast
                del selector.broadcast
                # print('sending', str(selector))
                await ws.send(str(selector))
                if broadcast.actions:
                    broadcast = str(broadcast)
                    for socket in clients:
                        await socket.send(broadcast)

        except websockets.exceptions.ConnectionClosed:
            logger.info('Connection closed.')
            clients.remove(ws)

    server = websockets.serve(socket_handler, **ws_options)
    loop.run_until_complete(server)
    try:
        logger.info('Server started on ws://%s:%d', ws_options['host'], ws_options['port'])
        loop.run_forever()
    except KeyboardInterrupt:
        pass


def start_app(host='127.0.0.1', port=5000, debug=True, log_level=logging.DEBUG, **ws_options):
    """Start the portkey app running in the event loop. This function should run once and only once for a portkey app. 
    
    Args:
        host: The hostname to listen on. Set this to '0.0.0.0' to have the server available externally
        port: The port of the server
        debug: Whether to automatically reload the program when the code is modified
        log_level: The log level to be passed to Portkey logger
        **ws_options: Other options to be passed to `websockets.serve()`
    """
    starter_function = functools.partial(_start_app, host=host, port=port, log_level=log_level, **ws_options)
    if debug:
        run_with_reloader(starter_function)
    else:
        starter_function()



from setuptools import setup
import sys

if sys.version_info < (3, 6):
    sys.exit('ERROR: Portkey only supports Python 3.6 or above.')

setup(
    name='Portkey',
    version='0.1.2',
    packages=['portkey'],
    url='https://github.com/red8012/portkey',
    license='Apache License 2.0',
    author='Yu-Ann Chen',
    author_email='red8012@gmail.com',
    description='A Python framework for interacting with in-browser DOM via websockets',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.6'
    ],
    install_requires=['websockets>=3.2', 'werkzeug>=0.11']
)



import inspect

from .utilities import *


def on(selector: str, event: str, filter_selector: str=None, throttle=False, stop_propagation=False):
    """Register a coroutine function as a handler for the given event.
    Corresponding to `$(selector).on(event, filter_selector, ...)` in jQuery

    Args:
        selector: A jQuery selector expression
        event: One or more space-separated event types and optional namespaces, such as "click".
        filter_selector: a selector string to filter the descendants of the selected elements that trigger the event.
            If this argument is omitted, the event is always triggered when it reaches the selected element.
        throttle: Whether to discard events when Portkey is busy.
            Should set to true if the event is going to fire at very high frequency.
            If set to false, all events will be queued and handled.
        stop_propagation: Prevents the event from bubbling up the DOM tree, 
            preventing any parent handlers from being notified of the event.

    Returns: A decorator that returns the original function.

    """

    def decorator(func):
        assert inspect.iscoroutinefunction(func), \
            'Only a coroutine function (a function defined with an async def syntax)' \
            ' can be registered as an event handler.'
        handler_info = get_handler_info(func)
        # assert handler_info.handler is None, \
        #     'You should only register one event on a handler function.'
        handler_info._handler = func
        handler_info.handler = id(func)
        handler_info.selector = selector
        handler_info.event = event
        handler_info.filter_selector = filter_selector
        handler_info.throttle = throttle
        handler_info.stop_propagation = stop_propagation
        return func
    return decorator


def onclick(selector: str, filter_selector: str=None, throttle=False, stop_propagation=False):
    """A shorthand for `on(selector, 'click').

    Args:
        selector: a jQuery selector expression
        filter_selector: a selector string to filter the descendants of the selected elements that trigger the event.
            If this argument is omitted, the event is always triggered when it reaches the selected element.
        throttle: Whether to discard events when Portkey is busy.
            Should set to true if the event is going to fire at very high frequency.
            If set to false, all events will be queued and handled.
        stop_propagation: Prevents the event from bubbling up the DOM tree, 
            preventing any parent handlers from being notified of the event.

    Returns: A decorator that returns the original function.

    """
    return on(selector, 'click',
              filter_selector=filter_selector, throttle=throttle, stop_propagation=stop_propagation)


def ready():
    """Register a coroutine function to be triggered when the Portkey client is ready.

    Returns: A decorator that returns the original function.

    """
    return on('body', '_Portkey:ready')


def _make_ui_data_tuple(v):
    if isinstance(v, tuple):
        return (False,) + v
    if isinstance(v, list):
        return [True] + v
    if isinstance(v, str):
        return v
    raise AttributeError('The value must be a tuple, a list or a string.')


def with_ui_data(**d):
    """Specifies the attributes to be retrieved from the Portkey client. 

    Args:
        **d: Key: variable name; value: attribute(s) to get 
            If the value is a string, it will be evaluated directly in the client.
            If the value is a tuple, it will be evaluated as `$(first).second(third, ...)`.
            If the value is a list, it works similarly as the tuple case, but returns the attributes of 
            EVERY matched element as a list.
            For the tuple and list cases, if the first element is 'target' or 'currentTarget', it will be evaluated as
            `$(event.target)` or `$(event.currentTarget)` respectively.
            
    Returns: A decorator that returns the original function.

    """

    def decorator(func):
        handler_info = get_handler_info(func)
        assert handler_info.uiData is None, \
            'You should only register a set of remote attributes on a handler function.'
        handler_info.uiData = {k: _make_ui_data_tuple(v) for k, v in d.items()}
        return func
    return decorator
